/************************************************************************/
/* Digital clock using MELT-107T LCD, DS1307 RTC and DS18B20 termometer.*/
/* Copyright (c) Sergey B. Voinov Aug 2010.                             */
/************************************************************************/

#include <htc.h>
#include "i2c.h"

// PIC16F648A

// I/O config:
//  O  RA0-RA3 = LCD DB0-DB3
//  O  RA4 = Speaker control
//  I  RA5 = MCLR
//  O  RA6 = LCD A0
//  O  RA7 = LCD WR1
//  I  RB0 = Interrupt from RTC
//  O  RB1 = RTC SCL
//  IO RB2 = RTC SDA
//  IO RB3 = Temperature sensor DQ
//  I  RB4 = B4
//  I  RB5 = B3
//  I  RB6 = B2
//  I  RB7 = B1

// MCU config:
//	bit 13   = 1    - code protection off
//	bit 9-12 = 0000 - unimplemented (read as 0)
//	bit 8    = 1    - data EPROM code protection off
//	bit 7    = 0    - RB3 is digital I/O
//	bit 6    = 0    - BOR disabled
//	bit 5    = 0    - RA5/MCLR pin function is digital I/O
//	bit 4    = 1
//	bit 3    = 1    - PWRT disabled
//	bit 2    = 0    - WDT disabled
//	bit 4,0-1= 100  - internal oscillator
__CONFIG(0x2118);

#define BUF_SIZE					RTC_DATA_SIZE
#define SPEAKER_OUT					RA4

#define LCD_NUM_SEGEMNTS			10
#define LCD_UNBLOCK_ADDR			0x0F
#define LCD_DATA					PORTA
#define LCD_DB0						RA0
#define LCD_DB1						RA1
#define LCD_DB2						RA2
#define LCD_DB3						RA3
#define LCD_A0						RA6
#define LCD_WR						RA7
#define LCD_MINUS					LCD_G
#define LCD_DASH					LCD_MINUS

/*************************
LCD segments layout:
           a
          ----
       f |    | b
         | g  |
          ----
       e |    | c
         |    |
          ---- . h
           d
*************************/

#define LCD_A						0b00001000
#define LCD_B						0b00100000
#define LCD_C						0b01000000
#define LCD_D						0b00000100
#define LCD_E						0b00000010
#define LCD_F						0b10000000
#define LCD_G						0b00000001
#define LCD_H						0b00010000

#define RTC_ADDR					0b11010000
#define RTC_START_ADDR				0
#define RTC_DATA_SIZE				11
#define RTC_TIME_ADDR				RTC_START_ADDR
#define RTC_DATE_ADDR				(RTC_START_ADDR + 4)
#define RTC_ALARM_ADDR				(RTC_DATE_ADDR + 4)

#define DS18B20_PIN					RB3
#define DS18B20_TRIS				TRISB3
#define DS18B20_SKIPROM_CMD			0xCC
#define DS18B20_READSCRATCH_CMD		0xBE
#define DS18B20_WRITESCRATCH_CMD	0x4E
#define DS18B20_CONVERT_CMD			0x44
#define DS18B20_CONVERT_DELAY		100000
#define DS18B20_SCRATCH_SIZE		9

// 1-wire bus timings (us)
#define DS18B20_PREINIT_TIME		500
#define DS18B20_RESET_TIME			500
#define DS18B20_INIT_DELAY			60
#define DS18B20_TIMESLOT_TIME		60
#define DS18B20_WRITE1_TIME			5

#define BTN1						RB7
#define BTN2						RB6
#define BTN3						RB5
#define BTN4						RB4

// Number of Timer1 overflow cycles between set mode blinks
#define CYCLE_MODE_DELAY			100
// Number of Timer1 overflow cycles between set mode blinks
#define BLINK_DELAY					8
// Number of Timer1 overflow cycles between alarm beeps
#define ALARM_DELAY					4
// Number of Timer1 cycles between speaker On/Off - defines beep frequency
#define ALARM_BEEP_DELAY			20


typedef enum eMode
{
	modeTime,
	modeDate,
	modeTemp,
	modeAlarm
} eMode;


// Digit codes for LCD
static const unsigned char digits[10] = {
	0b11101110, // '0'
	0b01100000, // '1'
	0b00101111, // '2'
	0b01101101, // '3'
	0b11100001, // '4'
	0b11001101, // '5'
	0b11001111, // '6'
	0b01101000, // '7'
	0b11101111, // '8'
	0b11101101  // '9'
};

// Read buffer
static unsigned char read_buf[BUF_SIZE];
// Clock mode
eMode mode = modeTime;
// When 1 - cycle through the modes
unsigned char cycle_modes = 0;
// When 1 - set mode is on
unsigned char is_set_mode = 0;
// Time/date field to change
unsigned char set_field = 0;
// Blinking flag for set mode
unsigned char blink = 0;
// Increment/decrement for a field
int field_inc = 0;
// Set to 1 when the time is equal to alarm time and alarm is enabled
unsigned char is_alarm;


// Forward declarations
void init();
void mcu_init();
void error(unsigned char code);
void next_mode();

void lcd_init();
void lcd_unblock();
void lcd_set_db(unsigned char data);
void lcd_write(unsigned char isdata, unsigned char nibble);
void lcd_write_segment(unsigned char addr, unsigned char data);
void lcd_write_digit(unsigned char addr, unsigned char digit);
void lcd_clear(unsigned char num_segments);
void lcd_test();

void rtc_init();
void rtc_read();
void rtc_write(unsigned char addr, unsigned char val);

void ds18b20_low();
void ds18b20_high();
void ds18b20_pulse();
void ds18b20_init();
unsigned char ds18b20_reset();
void ds18b20_write(unsigned char byte);
unsigned char ds18b20_read();
void ds18b20_read_scratchpad();
void ds18b20_read_temperature();

void display_time();
void display_date();
void display_temperature();
void display_alarm();
void display_error(unsigned char code);

unsigned char bcd_change(unsigned char old_bcd, int increment, unsigned char min_val, unsigned char max_val);

		
void main()
{
	unsigned char cycle_mode_count = 0;
	unsigned char blink_count = 0;
	unsigned char alarm_count = 0;
	unsigned char alarm_beep = 0;
	unsigned char alarm_beep_count = 0;
	
	init();
	
	TMR1ON = 1;
	
	for (;;)
	{
		if (!TMR1IF)
		{
			if (alarm_beep)
			{
				// Handle alarm beep frequency
				if (++alarm_beep_count > ALARM_BEEP_DELAY)
				{
					SPEAKER_OUT ^= 1;
					alarm_beep_count = 0;	
				}
			}
			else
			{
				// Turn off speaker
				SPEAKER_OUT = 1;
			}	
			continue;
		}	
	
		TMR1IF = 0;

		if (cycle_modes)
		{
			// Handle mode changes
			if (++cycle_mode_count > ((mode == modeTemp) ? CYCLE_MODE_DELAY / 2 : CYCLE_MODE_DELAY))
			{
				next_mode();
				cycle_mode_count = 0;	
			}
		}
		else
		{
			cycle_mode_count = 0;
		}	

		// Handle blink intervals
		if (++blink_count > BLINK_DELAY)
		{
			blink ^= 1;
			blink_count = 0;	
		}

		if (is_alarm)
		{
			// Handle alarm beep intervals
			if (++alarm_count > ALARM_DELAY)
			{
				alarm_beep ^= 1;
				alarm_count = 0;	
			}
		}
		else
		{
			alarm_beep = 0;
		}
			
		// RTC is read about 15 times a second
		// to minimize power consumption
				
		switch (mode)
		{
			case modeTime:
				rtc_read();
				display_time();
				break;
				
			case modeDate:
				rtc_read();
				display_date();
				break;
				
			case modeTemp:
				ds18b20_read_temperature();
				display_temperature();
				rtc_read();
				break;
				
			case modeAlarm:
				rtc_read();
				display_alarm();
				break;
		}
		
		// Need to change a field
		if (is_set_mode && field_inc)
		{
			unsigned char addr = 0, new_val = 0;
			unsigned char min_val = 0, max_val = 0;
			
			switch (mode)
			{
				case modeTime:
					addr = RTC_TIME_ADDR + 2 - set_field;
					switch (set_field)
					{
						case 0: min_val = 0; max_val = 0x23; break; // Hours
						case 1:
						case 2: min_val = 0; max_val = 0x59; break; // Minutes, seconds
					}	
					break;
					
				case modeDate:
					addr = RTC_DATE_ADDR + set_field;
					switch (set_field)
					{
						case 0: min_val = 1; max_val = 0x31; break; // Day
						case 1: min_val = 1; max_val = 0x12; break; // Month
						case 2: min_val = 0; max_val = 0x99; break; // Year
					}	
					break;
					
				case modeAlarm:
					addr = RTC_ALARM_ADDR + 2 - set_field;
					switch (set_field)
					{
						case 0: min_val = 0; max_val = 0x01; break; // Minutes
						case 1: min_val = 0; max_val = 0x23; break; // Hours
						case 2: min_val = 0; max_val = 0x59; break; // On/Off
					}	
					break;
					
				default:
					continue;
			}
			
			if (mode == modeTime && set_field == 2)
			{
				new_val = 0; // Seconds are just being reset
			}
			else
			{	
				new_val = bcd_change(read_buf[addr], field_inc, min_val, max_val);
			}
			
			// Write new value to RTC
			rtc_write(addr, new_val);
			
			// Reset change indicator
			field_inc = 0;
		}
		
		// Check for alarm
		is_alarm = ((read_buf[RTC_ALARM_ADDR + 2] == 0x01) && // Alarm is On
			(read_buf[RTC_TIME_ADDR + 1] == read_buf[RTC_ALARM_ADDR + 0]) && // Minutes are equal
			((read_buf[RTC_TIME_ADDR + 2] & 0x3F) == read_buf[RTC_ALARM_ADDR + 1])) ? 1 : 0; // Hours are equal
	}	
}

void init()
{
	mcu_init();
	lcd_init();
	rtc_init();
	ds18b20_init();
}

void mcu_init()
{
	// Disable fuckin' comparators!
	CMCON = 0b00000111;
	
	// Configure port A
	TRISA = 0b00100000;
	PORTA = 0b00010000;
	
	// Configure port B
	TRISB = 0b11111101;
	PORTB = 0b11111111;
	
	// PORTB pull-ups are enabled by individual port latch values
	RBPU = 0;
	
	// Turn off speaker
	SPEAKER_OUT = 1;
	
	// Enable interrupts
	RBIE = 1;
	GIE = 1;
}

void error(unsigned char code)
{
	display_error(code);
	for (;;);
}

void next_mode()
{
	// Handle mode change Time->Date->Temperature->Time->...
	++mode;
	if (mode > (cycle_modes ? modeTemp : modeAlarm))
	{
		mode = modeTime;
	}
	// Deactivate Set mode
	is_set_mode = 0;
}	
	
void lcd_init()
{
	lcd_unblock();
	lcd_clear(LCD_NUM_SEGEMNTS);
}	

void lcd_unblock()
{
	lcd_write(0, LCD_UNBLOCK_ADDR);
	lcd_write(1, 1);
}
	
void lcd_write(unsigned char isdata, unsigned char nibble)
{
	LCD_DATA = (LCD_DATA & 0xF0) | (nibble & 0x0F);
	LCD_A0 = isdata;
	LCD_WR = 1;
	LCD_WR = 0;
}

void lcd_write_segment(unsigned char addr, unsigned char data)
{
	lcd_write(0, addr);
	lcd_write(1, data);
	lcd_write(1, data >> 4);
}

void lcd_write_digit(unsigned char addr, unsigned char digit)
{
	unsigned char flag = 0;
	
	if (digit & 0x80)
	{
		digit &= ~0x80;
		flag = LCD_H;	// Decimal point
	}
		
	if (digit <= 9)
	{
		lcd_write_segment(addr, digits[digit] | flag);
	}
}

void lcd_clear(unsigned char num_segments)
{
	unsigned char i = 0;
	for (; i < LCD_NUM_SEGEMNTS && i < num_segments; ++i)
	{
		lcd_write_segment(i, 0);
	}	
}
	
void lcd_test()
{
	unsigned char i = 0;
	for (; i < LCD_NUM_SEGEMNTS && i < 10; ++i)
	{
		lcd_write_digit(i, i);
	}	
}
	
void rtc_init()
{
	if (i2c_WriteTo(RTC_ADDR))
	{
		error(1);
	}
	else
	{
		i2c_Stop();
	}	
}	

void rtc_read()
{
	i2c_WriteTo(RTC_ADDR);
	i2c_PutByte(RTC_START_ADDR);
	i2c_Stop();
	i2c_ReadFrom(RTC_ADDR);
	i2c_GetString(read_buf, RTC_DATA_SIZE);
	i2c_Stop();
}

void rtc_write(unsigned char addr, unsigned char val)
{
	i2c_WriteTo(RTC_ADDR);
	i2c_PutByte(addr);
	i2c_PutByte(val);
	i2c_Stop();
}
	
void ds18b20_low()
{
	// Pull DQ line low
	DS18B20_PIN = 0;
	DS18B20_TRIS = 0;	
}
	
void ds18b20_high()
{
	// Release DQ line - high
	DS18B20_PIN = 1;
	DS18B20_TRIS = 1;
}	

void ds18b20_pulse()
{
	// Pulse DQ line
	DS18B20_TRIS = 0;	
	DS18B20_PIN = 0;
	DS18B20_PIN = 1;
	DS18B20_TRIS = 1;
}
	
void ds18b20_init()
{
	// Set idle condition for 1-wire bus
	ds18b20_high();
	__delay_us(DS18B20_PREINIT_TIME);

	// Check DS18B20 presence	
	if (ds18b20_reset())
	{
		error(2);
	}	

	// Set 9-bit resolution for "fast" conversion
	ds18b20_write(DS18B20_SKIPROM_CMD);
	ds18b20_write(DS18B20_WRITESCRATCH_CMD);
	ds18b20_write(0); // Th
	ds18b20_write(0); // Tl
	ds18b20_write(0b00011111); // Configuration register
}

// Returns 0 if DS18B20 presents, 1 otherwise
unsigned char ds18b20_reset()
{
	unsigned char presense;
	
	// Reset pulse - 480us minimum
	ds18b20_low();
	__delay_us(DS18B20_RESET_TIME);
	
	// Release the bus
	ds18b20_high();
	
	// DS18B20 waits 15-60us
	__delay_us(DS18B20_INIT_DELAY);
	
	// DS18B20 transmits presence pulse for 60-240us
	presense = DS18B20_PIN;
	
	// Wait until DS18B20 releases the bus
	while (!DS18B20_PIN);
	
	return presense;
}

void ds18b20_write(unsigned char byte)
{
	unsigned char i = 0;
	
	for (; i < 8; ++i)
	{
		// Issue write time slot
		if (byte & 1)
		{
			ds18b20_low();
			// Write "1" slot
			__delay_us(DS18B20_WRITE1_TIME);
			ds18b20_high();
			// Wait till the end of the timeslot
			__delay_us(DS18B20_TIMESLOT_TIME - DS18B20_WRITE1_TIME);
		}
		else
		{
			ds18b20_low();
			// Write "0" slot
			__delay_us(DS18B20_TIMESLOT_TIME);
			ds18b20_high();
		}
		byte >>= 1;
	}	
}
	
unsigned char ds18b20_read()
{
	unsigned char byte = 0, i = 0;
	
	for (; i < 8; ++i)
	{
		byte >>= 1;
		// Issue read time slot
		ds18b20_pulse();
		// Wait for samling time (15us maximum )
		//__delay_us(10);
		if (DS18B20_PIN)	// "1"
		{
			byte |= 0x80;
		}
		// Wait till the end of the timeslot
		__delay_us(DS18B20_TIMESLOT_TIME);
	}
	
	return byte;	
}

void ds18b20_read_scratchpad()
{
	unsigned char i = 0;
	
	ds18b20_reset();
	ds18b20_write(DS18B20_SKIPROM_CMD);
	ds18b20_write(DS18B20_READSCRATCH_CMD);
	
	for (; i < DS18B20_SCRATCH_SIZE; ++i)
	{
		read_buf[i] = ds18b20_read();
	}
}

void ds18b20_read_temperature()
{
	ds18b20_reset();
	ds18b20_write(DS18B20_SKIPROM_CMD);
	ds18b20_write(DS18B20_CONVERT_CMD);
	__delay_us(DS18B20_CONVERT_DELAY);
	ds18b20_read_scratchpad();
}

void display_time()
{
	// Write seconds
	if (!is_set_mode || !(set_field == 2 && blink))
	{
		lcd_write_digit(LCD_NUM_SEGEMNTS - 1, read_buf[RTC_TIME_ADDR + 0] & 0x0F);
		lcd_write_digit(LCD_NUM_SEGEMNTS - 2, (read_buf[RTC_TIME_ADDR + 0] >> 4) & 0x07);
	}
	else
	{
		lcd_write_segment(LCD_NUM_SEGEMNTS - 1, 0);
		lcd_write_segment(LCD_NUM_SEGEMNTS - 2, 0);
	}	
	lcd_write_segment(LCD_NUM_SEGEMNTS - 3, LCD_DASH);

	// Write minutes
	if (!is_set_mode || !(set_field == 1 && blink))
	{
		lcd_write_digit(LCD_NUM_SEGEMNTS - 4, read_buf[RTC_TIME_ADDR + 1] & 0x0F);
		lcd_write_digit(LCD_NUM_SEGEMNTS - 5, read_buf[RTC_TIME_ADDR + 1] >> 4);
	}
	else
	{
		lcd_write_segment(LCD_NUM_SEGEMNTS - 4, 0);
		lcd_write_segment(LCD_NUM_SEGEMNTS - 5, 0);
	}	
	lcd_write_segment(LCD_NUM_SEGEMNTS - 6, LCD_DASH);
	
	// Write hours
	if (!is_set_mode || !(set_field == 0 && blink))
	{
		lcd_write_digit(LCD_NUM_SEGEMNTS - 7, read_buf[RTC_TIME_ADDR + 2] & 0x0F);
		lcd_write_digit(LCD_NUM_SEGEMNTS - 8, (read_buf[RTC_TIME_ADDR + 2] >> 4) & 0x03);
	}	
	else
	{
		lcd_write_segment(LCD_NUM_SEGEMNTS - 7, 0);
		lcd_write_segment(LCD_NUM_SEGEMNTS - 8, 0);
	}
		
	lcd_write_segment(LCD_NUM_SEGEMNTS - 9, 0);
	
	// Write alarm indicator
	if (is_alarm && blink)
	{
		lcd_write_segment(LCD_NUM_SEGEMNTS - 10, LCD_A | LCD_B | LCD_C | LCD_E | LCD_F | LCD_G); // 'A'
	}
	else
	{
		lcd_write_segment(LCD_NUM_SEGEMNTS - 10, 0);
	}		
}

void display_date()
{
	// Write year - 2000+
	if (!is_set_mode || !(set_field == 2 && blink))
	{	
		lcd_write_digit(LCD_NUM_SEGEMNTS - 1, read_buf[RTC_DATE_ADDR + 2] & 0x0F);
		lcd_write_digit(LCD_NUM_SEGEMNTS - 2, (read_buf[RTC_DATE_ADDR + 2] >> 4) & 0x0F);
		lcd_write_digit(LCD_NUM_SEGEMNTS - 3, 0);
		lcd_write_digit(LCD_NUM_SEGEMNTS - 4, 2);
	}
	else
	{
		lcd_write_segment(LCD_NUM_SEGEMNTS - 1, 0);
		lcd_write_segment(LCD_NUM_SEGEMNTS - 2, 0);
		lcd_write_segment(LCD_NUM_SEGEMNTS - 3, 0);
		lcd_write_segment(LCD_NUM_SEGEMNTS - 4, 0);
	}		
	lcd_write_segment(LCD_NUM_SEGEMNTS - 5, LCD_DASH);

	// Write month
	if (!is_set_mode || !(set_field == 1 && blink))
	{
		lcd_write_digit(LCD_NUM_SEGEMNTS - 6, read_buf[RTC_DATE_ADDR + 1] & 0x0F);
		lcd_write_digit(LCD_NUM_SEGEMNTS - 7, read_buf[RTC_DATE_ADDR + 1] >> 4);
	}
	else
	{
		lcd_write_segment(LCD_NUM_SEGEMNTS - 6, 0);
		lcd_write_segment(LCD_NUM_SEGEMNTS - 7, 0);
	}	
	lcd_write_segment(LCD_NUM_SEGEMNTS - 8, LCD_DASH);
	
	// Write day
	if (!is_set_mode || !(set_field == 0 && blink))
	{
		lcd_write_digit(LCD_NUM_SEGEMNTS - 9, read_buf[RTC_DATE_ADDR + 0] & 0x0F);
		lcd_write_digit(LCD_NUM_SEGEMNTS - 10, read_buf[RTC_DATE_ADDR + 0] >> 4);
	}	
	else
	{
		lcd_write_segment(LCD_NUM_SEGEMNTS - 9, 0);
		lcd_write_segment(LCD_NUM_SEGEMNTS - 10, 0);
	}	
}	

void display_temperature()
{
	unsigned char temp = (read_buf[1] << 4) | ((read_buf[0] >> 4) & 0x0F);
	unsigned char minus = 0, i = LCD_NUM_SEGEMNTS;
	
	if (temp & 0x80)
	{
		temp = -temp;
		minus = LCD_MINUS;
	}
	
	lcd_write_segment(--i, LCD_A | LCD_D | LCD_E | LCD_F); // 'C'
	lcd_write_segment(--i, LCD_A | LCD_B | LCD_F | LCD_G); // 'o'
	
	lcd_write_digit(--i, temp % 10);
	for (temp /= 10; temp; temp /= 10)
	{
		lcd_write_digit(--i, temp % 10);
	}
		
	lcd_write_segment(--i, minus);		
	lcd_clear(i);
	
	// Write alarm indicator
	if (is_alarm && blink)
	{
		lcd_write_segment(LCD_NUM_SEGEMNTS - 10, LCD_A | LCD_B | LCD_C | LCD_E | LCD_F | LCD_G); // 'A'
	}
	else
	{
		lcd_write_segment(LCD_NUM_SEGEMNTS - 10, 0);
	}	
}

void display_alarm()
{
	// Write minutes
	if (!is_set_mode || !(set_field == 2 && blink))
	{
		lcd_write_digit(LCD_NUM_SEGEMNTS - 1, read_buf[RTC_ALARM_ADDR + 0] & 0x0F);
		lcd_write_digit(LCD_NUM_SEGEMNTS - 2, read_buf[RTC_ALARM_ADDR + 0] >> 4);
	}
	else
	{
		lcd_write_segment(LCD_NUM_SEGEMNTS - 1, 0);
		lcd_write_segment(LCD_NUM_SEGEMNTS - 2, 0);
	}	
	lcd_write_segment(LCD_NUM_SEGEMNTS - 3, LCD_DASH);
	
	// Write hours
	if (!is_set_mode || !(set_field == 1 && blink))
	{
		lcd_write_digit(LCD_NUM_SEGEMNTS - 4, read_buf[RTC_ALARM_ADDR + 1] & 0x0F);
		lcd_write_digit(LCD_NUM_SEGEMNTS - 5, (read_buf[RTC_ALARM_ADDR + 1] >> 4) & 0x03);
	}	
	else
	{
		lcd_write_segment(LCD_NUM_SEGEMNTS - 4, 0);
		lcd_write_segment(LCD_NUM_SEGEMNTS - 5, 0);
	}
	
	lcd_write_segment(LCD_NUM_SEGEMNTS - 6, 0);
	
	// Write On/Off indicator
	if (!is_set_mode || !(set_field == 0 && blink))
	{
		lcd_write_segment(LCD_NUM_SEGEMNTS - 7, (read_buf[RTC_ALARM_ADDR + 2] == 0x01) ?
			(LCD_B | LCD_C | LCD_E | LCD_F | LCD_G) : // 'H'
			(LCD_A | LCD_F | LCD_E | LCD_G)           // 'F'
		);
		lcd_write_segment(LCD_NUM_SEGEMNTS - 8, LCD_A | LCD_F | LCD_E | LCD_B | LCD_C | LCD_D); // 'O'
	}	
	else
	{
		lcd_write_segment(LCD_NUM_SEGEMNTS - 7, 0);
		lcd_write_segment(LCD_NUM_SEGEMNTS - 8, 0);
	}
	
	lcd_write_segment(LCD_NUM_SEGEMNTS - 9, 0);
	lcd_write_segment(LCD_NUM_SEGEMNTS - 10, LCD_A | LCD_B | LCD_C | LCD_E | LCD_F | LCD_G); // 'A'
}

void display_error(unsigned char code)
{
	unsigned char i = LCD_NUM_SEGEMNTS;
	
	lcd_write_digit(--i, code);
	lcd_write_segment(--i, LCD_A | LCD_F | LCD_E);                         // 'R'
	lcd_write_segment(--i, LCD_A | LCD_F | LCD_E | LCD_B | LCD_C | LCD_D); // 'O'
	lcd_write_segment(--i, LCD_A | LCD_F | LCD_E);                         // 'R'
	lcd_write_segment(--i, LCD_A | LCD_F | LCD_E);                         // 'R'
	lcd_write_segment(--i, LCD_A | LCD_F | LCD_E | LCD_D | LCD_G);         // 'E'
	
	lcd_clear(i);
}


unsigned char old_btn1 = 0;
unsigned char old_btn2 = 0;
unsigned char old_btn3 = 0;
unsigned char old_btn4 = 0;

static void interrupt
isr(void)
{
	if (RBIF)
	{
		unsigned char new_btn1 = BTN1;
		unsigned char new_btn2 = BTN2;
		unsigned char new_btn3 = BTN3;
		unsigned char new_btn4 = BTN4;
		
		RBIF = 0;
		
		if (old_btn1 && !new_btn1) // Button1 pressed
		{
			if (cycle_modes)
			{
				cycle_modes = 0;
			}	
			
			next_mode();
			
			if (mode == modeTime)
			{
				cycle_modes = 1;
			}	
		}	

		if (old_btn2 && !new_btn2) // Button2 pressed
		{
			// Handle Set mode
			if (!is_set_mode && (mode != modeTemp))
			{
				// Can only be activated for time, date and temperature
				is_set_mode = 1;
				set_field = 0;
				cycle_modes = 0;
			}
			else
			{
				// Cycle through the fields
				++set_field;
				
				// Switch it off after last field
				if (mode == modeTime && set_field > 2)
				{
					is_set_mode = 0;
				}
				else if (mode == modeDate && set_field > 2)
				{
					is_set_mode = 0;
				}
				else if (mode == modeAlarm && set_field > 2)
				{
					is_set_mode = 0;
				}	
			}	
		}	

		if (old_btn3 && !new_btn3) // Button3 pressed
		{
			// Increment field
			if (is_set_mode && !field_inc)
			{
				field_inc = +1;
			}	
		}	
		
		if (old_btn4 && !new_btn4) // Button4 pressed
		{
			// Decrement field
			if (is_set_mode && !field_inc)
			{
				field_inc = -1;
			}	
		}	

		old_btn1 = new_btn1;
		old_btn2 = new_btn2;
		old_btn3 = new_btn3;
		old_btn4 = new_btn4;
	}	
}

// Used to change (+-1) correct BCD value, returns new BCD value.
unsigned char bcd_change(unsigned char old_bcd, int increment, unsigned char min_val, unsigned char max_val)
{
	unsigned char new_bcd = old_bcd + increment;
	
	if ((new_bcd & 0x0F) == 10)
	{
		// Correct for increment
		new_bcd += 6;
	}
	else if ((new_bcd & 0x0F) == 0x0F)
	{
		// Correct for decrement
		new_bcd -= 6;
	}
	
	if ((new_bcd & 0xF0) == 0xF0)
	{
		// Correct for underflow
		new_bcd = max_val;
	}
	
	if (new_bcd > max_val)
	{
		new_bcd = min_val;
	}
	else if (new_bcd < min_val)
	{
		new_bcd = max_val;
	}
	
	return new_bcd;	
}
